@@ -87,19 +87,6 @@ static inline void drbd_set_my_capacity(struct drbd_conf *mdev,
 
 #define drbd_bio_uptodate(bio) bio_flagged(bio, BIO_UPTODATE)
 
-static inline int drbd_bio_has_active_page(struct bio *bio)
-{
-	struct bio_vec *bvec;
-	int i;
-
-	__bio_for_each_segment(bvec, bio, i, 0) {
-		if (page_count(bvec->bv_page) > 1)
-			return 1;
-	}
-
-	return 0;
-}
-
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
 /* Before Linux-2.6.24 bie_endio() had the size of the bio as second argument.
    See 6712ecf8f648118c3363c142196418f89a510b90 */
@@ -117,8 +104,7 @@ static inline int drbd_bio_has_active_page(struct bio *bio)
 
 /* bi_end_io handlers */
 extern BIO_ENDIO_TYPE drbd_md_io_complete BIO_ENDIO_ARGS(struct bio *bio, int error);
-extern BIO_ENDIO_TYPE drbd_endio_read_sec BIO_ENDIO_ARGS(struct bio *bio, int error);
-extern BIO_ENDIO_TYPE drbd_endio_write_sec BIO_ENDIO_ARGS(struct bio *bio, int error);
+extern BIO_ENDIO_TYPE drbd_endio_sec BIO_ENDIO_ARGS(struct bio *bio, int error);
 extern BIO_ENDIO_TYPE drbd_endio_pri BIO_ENDIO_ARGS(struct bio *bio, int error);
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
@@ -651,14 +637,16 @@ typedef unsigned gfp_t;
  * this "backport" does not close the race that lead to the API change,
  * but only provides an equivalent function call.
  */
-#ifndef KERNEL_HAS_PROC_CREATE
-static inline struct proc_dir_entry *proc_create(const char *name,
+#ifndef KERNEL_HAS_PROC_CREATE_DATA
+static inline struct proc_dir_entry *proc_create_data(const char *name,
 	mode_t mode, struct proc_dir_entry *parent,
-	struct file_operations *proc_fops)
+	struct file_operations *proc_fops, void *data)
 {
 	struct proc_dir_entry *pde = create_proc_entry(name, mode, parent);
-	if (pde)
+	if (pde) {
 		pde->proc_fops = proc_fops;
+		pde->data = data;
+	}
 	return pde;
 }
 
@@ -689,5 +677,57 @@ static inline struct proc_dir_entry *proc_create(const char *name,
 
 #endif
 
+#ifdef NEED_BLK_QUEUE_MAX_HW_SECTORS
+static inline void blk_queue_max_hw_sectors(struct request_queue *q, unsigned int max)
+{
+	blk_queue_max_sectors(q, max);
+}
+#endif
+
+#ifdef NEED_BLK_QUEUE_MAX_SEGMENTS
+static inline void blk_queue_max_segments(struct request_queue *q, unsigned short max_segments)
+{
+	blk_queue_max_phys_segments(q, max_segments);
+	blk_queue_max_hw_segments(q, max_segments);
+#define BLK_MAX_SEGMENTS MAX_HW_SEGMENTS /* or max MAX_PHYS_SEGMENTS. Probably does not matter */
+}
+#endif
+
+#ifdef NEED_ATOMIC_ADD_UNLESS
+#ifndef atomic_cmpxchg
+static inline int atomic_cmpxchg(atomic_t *v, int old, int new)
+{
+	return cmpxchg(&v->counter, old, new);
+}
+#endif
+
+/**
+ * atomic_add_unless - add unless the number is already a given value
+ * @v: pointer of type atomic_t
+ * @a: the amount to add to v...
+ * @u: ...unless v is equal to u.
+ *
+ * Atomically adds @a to @v, so long as @v was not already @u.
+ * Returns non-zero if @v was not @u, and zero otherwise.
+ */
+static inline int atomic_add_unless(atomic_t *v, int a, int u)
+{
+	int c, old;
+	c = atomic_read(v);
+	for (;;) {
+		if (unlikely(c == (u)))
+			break;
+		old = atomic_cmpxchg((v), c, c + (a));
+		if (likely(old == c))
+			break;
+		c = old;
+	}
+	return c != (u);
+}
+#endif
+
+#ifdef NEED_BOOL_TYPE
+typedef _Bool                   bool;
+#endif
 
 #endif
