@@ -34,22 +34,35 @@ test -n "$KDIR"
 # ok, now we have a KDIR; cd into it, in case we detect relative pathes
 pushd $KDIR
 
-KDIR=${KDIR%/}
+KDIR=$(pwd)
 if test -z "$O"; then
 	## just in case...
 	## detect if $KDIR points to something which is actually $O ...
-	X=$( make no-such-makefile-target 2>/dev/null |
-	     sed -ne '/ -C .* O=.* no-such-makefile-target$/p' | tr -s ' ' )
-	if [[ -n $X ]]; then
+	## If someone _please_ make this easier :(
+	X=$( make no-such-makefile-target V=1 2>/dev/null |
+	     sed -ne '/ -C .* O=.* no-such-makefile-target$/p' \
+		-e 's/^[[:space:]]*\(KBUILD_SRC=[^ ]\+\).*$/\1/p' |
+		tr -s '\t ' '  ')
+	case $X in
+	KBUILD_SRC=*)
+		O=$KDIR
+		KDIR=${X#KBUILD_SRC=}
+		;;
+	*' -C '*)
 		KDIR=${X##* -C }; KDIR=${KDIR%% *}; KDIR=$(cd $KDIR && pwd)
 		O=${X##* O=}; O=${O%% *}; O=$(cd $KDIR && cd $O && pwd)
-	else
-		O=$KDIR;
-	fi
+		;;
+	*)	
+		O=$KDIR ;;
+	esac
 else
 	O=${O%/}
+	test -d $O
 fi
 
+echo "KDIR=$KDIR"
+echo "O=$O"
+
 # some paranoia: check that all files are where we expect them
 ls > /dev/null \
 $KDIR/{Makefile,include/linux/{gfp,types,slab,net}.h}
@@ -70,6 +83,7 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   fi
   # stupid vendor kernels grrr...
   have_atomic_add=0
+  have_atomic_add_unless=0
   # btw, don't ask why I don't use grep -qs $a $b $c 
   # it simply does not work always...
   for f in $O/include/asm/atomic.h \
@@ -78,11 +92,14 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
     $O/include/asm/atomic_32.h \
     $O/include2/asm/atomic_32.h \
     $O/include/asm/arch/atomic_32.h \
-    $KDIR/include/asm-generic/atomic.h
+    $KDIR/include/asm-generic/atomic.h \
+    $KDIR/arch/x86/include/asm/atomic_32.h # Assume ARCHs are in sync, feature wise.
   do
     if grep_q "atomic_add_return" $f; then
       have_atomic_add=1
-      break
+    fi
+    if grep_q "atomic_add_unless" $f; then
+      have_atomic_add_unless=1
     fi
   done
   if grep_q "typedef.*kmem_cache_s" $KDIR/include/linux/slab.h ; then
@@ -136,10 +153,10 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   else
     have_linux_byteorder_swabb_h=0
   fi
-  if grep_q "proc_create(" $KDIR/include/linux/proc_fs.h ; then
-    have_proc_create=1
+  if grep_q "proc_create_data(" $KDIR/include/linux/proc_fs.h ; then
+    have_proc_create_data=1
   else
-    have_proc_create=0
+    have_proc_create_data=0
   fi
   if grep_q "set_cpus_allowed_ptr(" $KDIR/include/linux/sched.h ; then
     have_set_cpus_allowed_ptr=1
@@ -151,6 +168,21 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   else
     have_netlink_skb_parms=0
   fi
+  if grep_q "blk_queue_max_hw_sectors" $KDIR/include/linux/blkdev.h ; then
+    need_blk_queue_max_hw_sectors=0
+  else
+    need_blk_queue_max_hw_sectors=1
+  fi
+  if grep_q "blk_queue_max_segments" $KDIR/include/linux/blkdev.h ; then
+    need_blk_queue_max_segments=0
+  else
+    need_blk_queue_max_segments=1
+  fi
+  if grep_q "typedef.*bool" $KDIR/include/linux/types.h ; then
+    have_bool_type=1
+  else
+    have_bool_type=0
+  fi
 else
     # not a 2.6. kernel. just leave it alone...
     exit 0
@@ -186,12 +218,20 @@ perl -pe "
   { ( $have_kvec ? '' : '//' ) . \$1}e;
  s{.*(#define HAVE_LINUX_BYTEORDER_SWABB_H.*)}
   { ( $have_linux_byteorder_swabb_h ? '' : '//' ) . \$1}e;
- s{.*(#define KERNEL_HAS_PROC_CREATE.*)}
-  { ( $have_proc_create ? '' : '//' ) . \$1}e;
+ s{.*(#define KERNEL_HAS_PROC_CREATE_DATA.*)}
+  { ( $have_proc_create_data ? '' : '//' ) . \$1}e;
  s{.*(#define HAVE_SET_CPUS_ALLOWED_PTR.*)}
   { ( $have_set_cpus_allowed_ptr ? '' : '//' ) . \$1}e;
  s{.*(#define KERNEL_HAS_CN_SKB_PARMS.*)}
   { ( $have_netlink_skb_parms ? '' : '//' ) . \$1}e;
+ s{.*(#define NEED_BLK_QUEUE_MAX_HW_SECTORS.*)}
+  { ( $need_blk_queue_max_hw_sectors ? '' : '//' ) . \$1}e;
+ s{.*(#define NEED_BLK_QUEUE_MAX_SEGMENTS.*)}
+  { ( $need_blk_queue_max_segments ? '' : '//' ) . \$1}e;
+ s{.*(#define NEED_ATOMIC_ADD_UNLESS.*)}
+  { ( $have_atomic_add_unless ? '//' : '' ) . \$1}e;
+ s{.*(#define NEED_BOOL_TYPE.*)}
+  { ( $have_bool_type ? '//' : '' ) . \$1}e;
  " \
 	  < ./linux/drbd_config.h \
 	  > ./linux/drbd_config.h.new
