@@ -153,6 +153,7 @@ static int adm_khelper(struct d_resource *, const char *);
 static int adm_generic_b(struct d_resource *, const char *);
 static int hidden_cmds(struct d_resource *, const char *);
 static int adm_outdate(struct d_resource *, const char *);
+static int adm_chk_resize(struct d_resource *res, const char *cmd);
 
 static char *get_opt_val(struct d_option *, const char *, char *);
 static void register_config_file(struct d_resource *res, const char *cfname);
@@ -332,6 +333,7 @@ struct adm_cmd cmds[] = {
 	{"fence-peer", adm_khelper, DRBD_acf3_handler},
 	{"local-io-error", adm_khelper, DRBD_acf3_handler},
 	{"pri-lost", adm_khelper, DRBD_acf3_handler},
+	{"initial-split-brain", adm_khelper, DRBD_acf3_handler},
 	{"split-brain", adm_khelper, DRBD_acf3_handler},
 	{"out-of-sync", adm_khelper, DRBD_acf3_handler},
 
@@ -339,6 +341,7 @@ struct adm_cmd cmds[] = {
 	{"resume-io", adm_generic_s, DRBD_acf4_advanced},
 	{"set-gi", admm_generic, DRBD_acf4_advanced},
 	{"new-current-uuid", adm_generic_s, DRBD_acf4_advanced},
+	{"check-resize", adm_chk_resize, DRBD_acf4_advanced},
 };
 
 
@@ -370,16 +373,16 @@ void schedule_dcmd(int (*function) (struct d_resource *, const char *),
 	deferred_cmds_tail[order] = d;
 }
 
-static int adm_generic(struct d_resource *res, const char *cmd, int flags);
+static void _adm_generic(struct d_resource *res, const char *cmd, int flags, pid_t *pid, int *fd, int *ex);
 
 /* Returns non-zero if the resource is down. */
 static int test_if_resource_is_down(struct d_resource *res)
 {
 	char buf[1024];
-	char *line;
+	int rr, s = 0;
 	int fd;
+	pid_t pid;
 	int old_verbose = verbose;
-	FILE *f;
 
 	if (dry_run) {
 		fprintf(stderr, "Logic bug: should not be dry-running here.\n");
@@ -387,7 +390,8 @@ static int test_if_resource_is_down(struct d_resource *res)
 	}
 	if (verbose == 1)
 		verbose = 0;
-	fd = adm_generic(res, "role", RETURN_STDOUT_FD | SUPRESS_STDERR);
+	_adm_generic(res, "role", SLEEPS_SHORT | RETURN_STDOUT_FD | SUPRESS_STDERR,
+			&pid, &fd, NULL);
 	verbose = old_verbose;
 
 	if (fd < 0) {
@@ -395,25 +399,18 @@ static int test_if_resource_is_down(struct d_resource *res)
 		exit(E_thinko);
 	}
 
-	f = fdopen(fd, "r");
-	if (f == NULL) {
-		fprintf(stderr, "Bad: could not fdopen file descriptor.\n");
-		exit(E_thinko);
+	while (1) {
+		rr = read(fd, buf + s, sizeof(buf) - s);
+		if (rr <= 0)
+			break;
+		s += rr;
 	}
+	close(fd);
 
-	errno = 0;
-	line = fgets(buf, sizeof(buf) - 1, f);
-	if (line == NULL && errno != 0) {
-		fprintf(stderr,
-			"Bad: fgets returned NULL while waiting for data: %m\n");
-		exit(E_thinko);
-	}
-	fclose(f);
+	waitpid(pid, NULL, 0);	/* Reap the child process, do not leave a zombie around. */
+	alarm(0);
 
-	waitpid(0, NULL, WNOHANG);	/* Reap the child process, do not leave a zombie around. */
-
-	if (line == NULL
-	    || strncmp(line, "Unconfigured", strlen("Unconfigured")) == 0)
+	if (s == 0 || strncmp(buf, "Unconfigured", strlen("Unconfigured")) == 0)
 		return 1;
 
 	return 0;
@@ -1206,7 +1203,7 @@ static void alarm_handler(int __attribute((unused)) signo)
 	alarm_raised = 1;
 }
 
-pid_t m_system(char **argv, int flags, struct d_resource *res)
+void m__system(char **argv, int flags, struct d_resource *res, pid_t *kid, int *fd, int *ex)
 {
 	pid_t pid;
 	int status, rv = -1;
@@ -1228,8 +1225,15 @@ pid_t m_system(char **argv, int flags, struct d_resource *res)
 			printf("%s ", shell_escape(*cmdline++));
 		}
 		printf("\n");
-		if (dry_run)
-			return 0;
+		if (dry_run) {
+			if (kid)
+				*kid = -1;
+			if (fd)
+				*fd = 0;
+			if (ex)
+				*ex = 0;
+			return;
+		}
 	}
 
 	/* flush stdout and stderr, so output of drbdadm
@@ -1290,12 +1294,14 @@ pid_t m_system(char **argv, int flags, struct d_resource *res)
 		alarm(timeout);
 	}
 
-	if (flags == RETURN_PID) {
-		return pid;
-	}
+	if (kid)
+		*kid = pid;
+	if (fd)
+		*fd = pipe_fds[0];
 
-	if (flags & (RETURN_STDOUT_FD | RETURN_STDERR_FD))
-		return pipe_fds[0];
+	if (flags & (RETURN_STDOUT_FD | RETURN_STDERR_FD)
+	||  flags == RETURN_PID)
+		return;
 
 	while (1) {
 		if (waitpid(pid, &status, 0) == -1) {
@@ -1342,7 +1348,8 @@ pid_t m_system(char **argv, int flags, struct d_resource *res)
 	fflush(stdout);
 	fflush(stderr);
 
-	return rv;
+	if (ex)
+		*ex = rv;
 }
 
 #define NA(ARGC) \
@@ -1391,7 +1398,7 @@ int adm_attach(struct d_resource *res, const char *unused __attribute((unused)))
 	make_options(opt);
 	argv[NA(argc)] = 0;
 
-	return m_system(argv, SLEEPS_LONG, res);
+	return m_system_ex(argv, SLEEPS_LONG, res);
 }
 
 struct d_option *find_opt(struct d_option *base, char *name)
@@ -1405,11 +1412,13 @@ struct d_option *find_opt(struct d_option *base, char *name)
 	return 0;
 }
 
-int adm_resize(struct d_resource *res, const char *unused __attribute((unused)))
+int adm_resize(struct d_resource *res, const char *cmd)
 {
 	char *argv[MAX_ARGS];
 	struct d_option *opt;
 	int i, argc = 0;
+	int silent;
+	int ex;
 
 	argv[NA(argc)] = drbdsetup;
 	ssprintf(argv[NA(argc)], "%d", res->me->device_minor);
@@ -1417,12 +1426,32 @@ int adm_resize(struct d_resource *res, const char *unused __attribute((unused)))
 	opt = find_opt(res->disk_options, "size");
 	if (opt)
 		ssprintf(argv[NA(argc)], "--%s=%s", opt->name, opt->value);
-	for (i = 0; i < soi; i++) {
+	for (i = 0; i < soi; i++)
 		argv[NA(argc)] = setup_opts[i];
-	}
 	argv[NA(argc)] = 0;
 
-	return m_system(argv, SLEEPS_SHORT, res);
+	/* if this is not "resize", but "check-resize", be silent! */
+	silent = strcmp(cmd, "resize") ? SUPRESS_STDERR : 0;
+	ex = m_system_ex(argv, SLEEPS_SHORT | silent, res);
+
+	if (ex)
+		return ex;
+
+	/* Record last-known bdev info.
+	 * Unfortunately drbdsetup did not have enough information
+	 * when doing the "resize", and in theory, _our_ information
+	 * about the backing device may even be wrong.
+	 * Call drbdsetup again, tell it to ask the kernel for
+	 * current config, and update the last known bdev info
+	 * according to that. */
+	/* argv[0] = drbdsetup;
+	 * argv[1] = minor; */
+	argv[2] = "check-resize";
+	argv[3] = NULL;
+	/* ignore exit code */
+	m_system_ex(argv, SLEEPS_SHORT | silent, res);
+
+	return 0;
 }
 
 int _admm_generic(struct d_resource *res, const char *cmd, int flags)
@@ -1454,7 +1483,7 @@ int _admm_generic(struct d_resource *res, const char *cmd, int flags)
 
 	argv[NA(argc)] = 0;
 
-	return m_system(argv, flags, res);
+	return m_system_ex(argv, flags, res);
 }
 
 static int admm_generic(struct d_resource *res, const char *cmd)
@@ -1462,7 +1491,7 @@ static int admm_generic(struct d_resource *res, const char *cmd)
 	return _admm_generic(res, cmd, SLEEPS_VERY_LONG);
 }
 
-static int adm_generic(struct d_resource *res, const char *cmd, int flags)
+static void _adm_generic(struct d_resource *res, const char *cmd, int flags, pid_t *pid, int *fd, int *ex)
 {
 	char *argv[MAX_ARGS];
 	int argc = 0, i;
@@ -1476,7 +1505,14 @@ static int adm_generic(struct d_resource *res, const char *cmd, int flags)
 	argv[NA(argc)] = 0;
 
 	setenv("DRBD_RESOURCE", res->name, 1);
-	return m_system(argv, flags, res);
+	m__system(argv, flags, res, pid, fd, ex);
+}
+
+static int adm_generic(struct d_resource *res, const char *cmd, int flags)
+{
+	int ex;
+	_adm_generic(res, cmd, flags, NULL, NULL, &ex);
+	return ex;
 }
 
 int adm_generic_s(struct d_resource *res, const char *cmd)
@@ -1569,12 +1605,26 @@ static int adm_outdate(struct d_resource *res, const char *cmd)
 	return rv;
 }
 
+/* shell equivalent:
+ * ( drbdsetup resize && drbdsetup check-resize ) || drbdmeta check-resize */
+static int adm_chk_resize(struct d_resource *res, const char *cmd)
+{
+	/* drbdsetup resize && drbdsetup check-resize */
+	int ex = adm_resize(res, cmd);
+	if (ex == 0)
+		return 0;
+
+	/* try drbdmeta check-resize */
+	return admm_generic(res, cmd);
+}
+
 static int adm_generic_b(struct d_resource *res, const char *cmd)
 {
 	char buffer[4096];
 	int fd, status, rv = 0, rr, s = 0;
+	pid_t pid;
 
-	fd = adm_generic(res, cmd, SLEEPS_SHORT | RETURN_STDERR_FD);
+	_adm_generic(res, cmd, SLEEPS_SHORT | RETURN_STDERR_FD, &pid, &fd, NULL);
 
 	if (fd < 0) {
 		fprintf(stderr, "Strange: got negative fd.\n");
@@ -1590,7 +1640,8 @@ static int adm_generic_b(struct d_resource *res, const char *cmd)
 		}
 
 		close(fd);
-		waitpid(0, &status, WNOHANG);
+		rr = waitpid(pid, &status, 0);
+		alarm(0);
 
 		if (WIFEXITED(status))
 			rv = WEXITSTATUS(status);
@@ -1599,7 +1650,12 @@ static int adm_generic_b(struct d_resource *res, const char *cmd)
 		}
 	}
 
-	if (rv == 11) {
+	/* see drbdsetup.c, print_config_error():
+	 *  11: some unspecific state change error
+	 *  17: SS_NO_UP_TO_DATE_DISK
+	 * In both cases, we don't need to retry with drbdmeta,
+	 * it would fail anyways with "Device is configured!" */
+	if (rv == 11 || rv == 17) {
 		/* Some state transition error, report it ... */
 		rr = write(fileno(stderr), buffer, s);
 		return rv;
@@ -1657,7 +1713,7 @@ static int adm_khelper(struct d_resource *res, const char *cmd)
 
 	if ((sh_cmd = get_opt_val(res->handlers, cmd, NULL))) {
 		argv[2] = sh_cmd;
-		rv = m_system(argv, SLEEPS_VERY_LONG, res);
+		rv = m_system_ex(argv, SLEEPS_VERY_LONG, res);
 	}
 	return rv;
 }
@@ -1721,7 +1777,7 @@ int adm_connect(struct d_resource *res,
 
 	argv[NA(argc)] = 0;
 
-	return m_system(argv, SLEEPS_SHORT, res);
+	return m_system_ex(argv, SLEEPS_SHORT, res);
 }
 
 struct d_resource *res_by_name(const char *name);
@@ -1752,7 +1808,7 @@ struct d_option *del_opt(struct d_option *base, struct d_option *item)
 // Need to convert after from resourcename to minor_number.
 void convert_after_option(struct d_resource *res)
 {
-	struct d_option *opt;
+	struct d_option *opt, *next;
 	struct d_resource *depends_on_res;
 
 	if (res == NULL)
@@ -1760,14 +1816,15 @@ void convert_after_option(struct d_resource *res)
 
 	opt = res->sync_options;
 	while ((opt = find_opt(opt, "after"))) {
+		next = opt->next;
 		depends_on_res = res_by_name(opt->value);
 		if (!depends_on_res || depends_on_res->ignore) {
 			res->sync_options = del_opt(res->sync_options, opt);
 		} else {
 			free(opt->value);
-			asprintf(&opt->value, "%d", depends_on_res->me->device_minor);
+			m_asprintf(&opt->value, "%d", depends_on_res->me->device_minor);
 		}
-		opt = opt->next;
+		opt = next;
 	}
 }
 
@@ -1801,6 +1858,15 @@ static int do_proxy(struct d_resource *res, int do_up)
 		exit(E_config_invalid);
 	}
 
+	if (!res->peer->proxy) {
+		fprintf(stderr,
+			"There is no proxy config for the peer in resource %s.\n",
+			res->name);
+		if (all_resources)
+			return 0;
+		exit(E_config_invalid);
+	}
+
 	argv[NA(argc)] = drbd_proxy_ctl;
 	argv[NA(argc)] = "-c";
 	if (do_up) {
@@ -1825,7 +1891,7 @@ static int do_proxy(struct d_resource *res, int do_up)
 	}
 	argv[NA(argc)] = 0;
 
-	rv = m_system(argv, SLEEPS_SHORT, res);
+	rv = m_system_ex(argv, SLEEPS_SHORT, res);
 	if (rv != 0)
 		return rv;
 
@@ -1846,7 +1912,7 @@ static int do_proxy(struct d_resource *res, int do_up)
 	}
 	argv[NA(argc)] = 0;
 	if (argc > 2)
-		return m_system(argv, SLEEPS_SHORT, res);
+		return m_system_ex(argv, SLEEPS_SHORT, res);
 	return rv;
 }
 
@@ -1883,7 +1949,7 @@ int adm_syncer(struct d_resource *res, const char *unused __attribute((unused)))
 
 	argv[NA(argc)] = 0;
 
-	return m_system(argv, SLEEPS_SHORT, res);
+	return m_system_ex(argv, SLEEPS_SHORT, res);
 }
 
 static int adm_up(struct d_resource *res,
@@ -1926,7 +1992,7 @@ static int adm_wait_c(struct d_resource *res,
 	}
 	argv[NA(argc)] = 0;
 
-	rv = m_system(argv, SLEEPS_FOREVER, res);
+	rv = m_system_ex(argv, SLEEPS_FOREVER, res);
 
 	return rv;
 }
@@ -2157,7 +2223,7 @@ static int adm_wait_ci(struct d_resource *ignored __attribute((unused)),
 		make_options(opt);
 		argv[NA(argc)] = 0;
 
-		pids[i++] = m_system(argv, RETURN_PID, res);
+		m__system(argv, RETURN_PID, res, &pids[i++], NULL, NULL);
 	}
 
 	wtime = global_options.dialog_refresh ? : -1;
@@ -2462,7 +2528,7 @@ int sanity_check_abs_cmd(char *cmd_name)
 		return 0;
 	}
 
-	if (!sb.st_mode & S_ISUID || sb.st_mode & S_IXOTH || sb.st_gid == 0) {
+	if (!(sb.st_mode & S_ISUID) || sb.st_mode & S_IXOTH || sb.st_gid == 0) {
 		static int did_header = 0;
 		if (!did_header)
 			fprintf(stderr,
@@ -2558,7 +2624,7 @@ void sanity_check_perm()
 
 void validate_resource(struct d_resource *res)
 {
-	struct d_option *opt;
+	struct d_option *opt, *next;
 	struct d_name *bpo;
 
 	if (!res->protocol) {
@@ -2572,15 +2638,29 @@ void validate_resource(struct d_resource *res)
 	} else {
 		res->protocol[0] = toupper(res->protocol[0]);
 	}
-	if ((opt = find_opt(res->sync_options, "after"))) {
+	/* there may be more than one "after" statement,
+	 * see commit 89cd0585 */
+	opt = res->sync_options;
+	while ((opt = find_opt(opt, "after"))) {
+		next = opt->next;
 		if (res_by_name(opt->value) == NULL) {
 			fprintf(stderr,
 				"%s:%d: in resource %s:\n\tresource '%s' mentioned in "
 				"'after' option is not known.\n",
 				res->config_file, res->start_line, res->name,
 				opt->value);
-			config_valid = 0;
+			/* Non-fatal if run from some script.
+			 * When deleting resources, it is an easily made
+			 * oversight to leave references to the deleted
+			 * resources in sync-after statements.  Don't fail on
+			 * every pacemaker-induced action, as it would
+			 * ultimately lead to all nodes committing suicide. */
+			if (no_tty)
+				res->sync_options = del_opt(res->sync_options, opt);
+			else
+				config_valid = 0;
 		}
+		opt = next;
 	}
 	if (res->ignore)
 		return;
@@ -2628,6 +2708,9 @@ void validate_resource(struct d_resource *res)
 		config_valid = 0;
 	}
 
+	if (!res->peer)
+		set_peer_in_resource(res, 0);
+
 	if (res->peer
 	    && ((res->me->proxy == NULL) != (res->peer->proxy == NULL))) {
 		fprintf(stderr,
