@@ -32,6 +32,7 @@
 #include <glob.h>
 #include <search.h>
 #include <fcntl.h>
+#include <unistd.h>
 
 #include "drbdadm.h"
 #include "linux/drbd_limits.h"
@@ -136,7 +137,10 @@ void m_strtoll_range(const char *s, char def_unit,
 		fprintf(stderr,
 			"%s:%d: %s %s => %llu%s out of range [%llu..%llu]%s.\n",
 			config_file, fline, name, s, r, unit, min, max, unit);
-		exit(E_config_invalid);
+		if (config_valid <= 1) {
+			config_valid = 0;
+			return;
+		}
 	}
 	if (DEBUG_RANGE_CHECK) {
 		fprintf(stderr,
@@ -228,6 +232,13 @@ void range_check(const enum range_checks what, const char *name,
 		m_strtoll_range(value, 1, name, DRBD_OUTDATED_WFC_TIMEOUT_MIN,
 				DRBD_OUTDATED_WFC_TIMEOUT_MAX);
 		break;
+
+	case R_DP_VOLUME:
+	case R_DP_INTERVAL:
+	case R_RS_THROTTLE_TH:
+	case R_RS_HOLD_OFF_TH:
+		/* ignored anyways; emit deprecation warning? */
+		break;
 	}
 }
 
@@ -1128,6 +1139,9 @@ void set_me_in_resource(struct d_resource* res, int match_on_proxy)
 		} else if (host->lower) {
 			if (!host->lower->me)
 				continue;
+		} else if (!host->on_hosts) {
+			/* huh? a resource without hosts to run on?! */
+			continue;
 		} else {
 			if (!name_in_names(nodeinfo.nodename, host->on_hosts) &&
 			    strcmp("_this_host", host->on_hosts->name))
@@ -1275,8 +1289,10 @@ void set_on_hosts_in_res(struct d_resource *res)
 			if (l_res == NULL) {
 				fprintf(stderr, "%s:%d: in resource %s, "
 					"referenced resource '%s' not defined.\n",
-					res->config_file, res->start_line, res->name, l_res->name);
+					res->config_file, res->start_line, res->name,
+					host->lower_name);
 				config_valid = 0;
+				continue;
 			}
 
 			/* Simple: host->on_hosts = concat_names(l_res->me->on_hosts, l_res->peer->on_hosts); */
@@ -1528,6 +1544,7 @@ void include_stmt(char *str)
 	int cwd_fd;
 	FILE *f;
 	size_t i;
+	int r;
 
 	/* in order to allow relative paths in include statements we change
 	   directory to the location of the current configuration file. */
@@ -1547,23 +1564,31 @@ void include_stmt(char *str)
 		exit(E_usage);
 	}
 
-	f = fopen(str, "r");
-	if (f) {
-		include_file(f, str);
-		fclose(f);
-	} else if (glob(str, 0, NULL, &glob_buf) == 0) {
+	r = glob(str, 0, NULL, &glob_buf);
+	if (r == 0) {
 		for (i=0; i<glob_buf.gl_pathc; i++) {
 			f = fopen(glob_buf.gl_pathv[i], "r");
-			if (f)
+			if (f) {
 				include_file(f, strdup(glob_buf.gl_pathv[i]));
-			fclose(f);
+				fclose(f);
+			} else {
+				fprintf(stderr,
+					"%s:%d: Failed to open include file '%s'.\n",
+					config_file, line, yylval.txt);
+				config_valid = 0;
+			}
 		}
 		globfree(&glob_buf);
+	} else if (r == GLOB_NOMATCH) {
+		if (!strchr(str, '?') && !strchr(str, '*') && !strchr(str, '[')) {
+			fprintf(stderr,
+				"%s:%d: Failed to open include file '%s'.\n",
+				config_file, line, yylval.txt);
+			config_valid = 0;
+		}
 	} else {
-		fprintf(stderr,
-			"%s:%d: Failed to open include file '%s'.\n",
-			config_file, line, yylval.txt);
-		config_valid = 0;
+		fprintf(stderr, "glob() failed: %d\n", r);
+		exit(E_usage);
 	}
 
 	if (fchdir(cwd_fd) < 0) {
