@@ -46,6 +46,10 @@ enum range_checks
 	R_WFC_TIMEOUT,
 	R_DEGR_WFC_TIMEOUT,
 	R_OUTDATED_WFC_TIMEOUT,
+	R_DP_VOLUME,
+	R_DP_INTERVAL,
+	R_RS_THROTTLE_TH,
+	R_RS_HOLD_OFF_TH,
 };
 
 enum yytokentype {
@@ -101,6 +105,7 @@ enum yytokentype {
 	TK_ERR_DQSTRING_TOO_LONG,
 	TK_ERR_DQSTRING,
 	TK_SCI,
+	TK_SDP,
 	TK_SSOCKS,
 	TK_IPV4,
 	TK_IPV6,
