@@ -182,6 +182,7 @@ static int sh_status_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *r
 static int cstate_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *rtl);
 static int dstate_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *rtl);
 static int uuids_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *rtl);
+static int lk_bdev_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *rtl);
 
 // convert functions for arguments
 static int conv_block_dev(struct drbd_argument *ad, struct drbd_tag_list *tl, char* arg);
@@ -202,6 +203,18 @@ static void show_handler(struct drbd_option *od, unsigned short* tp);
 static void show_bit(struct drbd_option *od, unsigned short* tp);
 static void show_string(struct drbd_option *od, unsigned short* tp);
 
+// ignore things that used to be options, but are no-ops now
+static int conv_ignore(struct drbd_option *od, struct drbd_tag_list *tl, char* arg)
+	{ return NO_ERROR; };
+#if 0
+static void show_ignore(struct drbd_option *od, unsigned short* tp)
+	{ };
+# endif
+static int usage_ignore(struct drbd_option *option, char* str, int strlen)
+	{ return 0; };
+static void usage_xml_ignore(struct drbd_option *option)
+	{ };
+
 // sub functions for events_cmd
 static int print_broadcast_events(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
 static int w_connected_state(unsigned int seq, int, struct drbd_nl_cfg_reply *reply);
@@ -261,6 +274,11 @@ struct option wait_cmds_options[] = {
 	conv_numeric, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
 		DRBD_ ## N ## _DEF ,U,UN  } }
+/* deprecated numeric, only present in "drbdsetup show", otherwise ignored. */
+#define DN(N,U,UN) \
+	conv_ignore, show_numeric, usage_ignore, usage_xml_ignore, \
+	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
+		DRBD_ ## N ## _DEF ,U,UN  } }
 #define EN_sndbuf(N,U,UN) \
 	conv_sndbuf, show_numeric, numeric_opt_usage, numeric_opt_xml, \
 	{ .numeric_param = { DRBD_ ## N ## _MIN, DRBD_ ## N ## _MAX, \
@@ -280,7 +298,8 @@ struct option wait_cmds_options[] = {
 struct drbd_cmd commands[] = {
 	{"primary", P_primary, F_CONFIG_CMD, {{ NULL,
 	 (struct drbd_option[]) {
-		 { "overwrite-data-of-peer",'o',T_overwrite_peer, EB   },
+		 { "overwrite-data-of-peer",'o',T_primary_force, EB   }, /* legacy name */
+		 { "force",'f',			T_primary_force, EB   },
 		 CLOSE_OPTIONS }} }, },
 
 	{"secondary", P_secondary, F_CONFIG_CMD, {{NULL, NULL}} },
@@ -333,6 +352,7 @@ struct drbd_cmd commands[] = {
 		 { "discard-my-data",'D', T_want_lose,     EB },
 		 { "data-integrity-alg",'d', T_integrity_alg,     ES },
 		 { "no-tcp-cork",'o',   T_no_cork,         EB },
+		 { "dry-run",'n',   T_dry_run,		   EB },
 		 CLOSE_OPTIONS }} }, },
 
 	{"disconnect", P_disconnect, F_CONFIG_CMD, {{NULL, NULL}} },
@@ -341,6 +361,7 @@ struct drbd_cmd commands[] = {
 	 (struct drbd_option[]) {
 		 { "size",'s',T_resize_size,		EN(DISK_SIZE_SECT,'s',"bytes") },
 		 { "assume-peer-has-space",'f',T_resize_force,	EB },
+		 { "assume-clean", 'c',        T_no_resync, EB },
 		 CLOSE_OPTIONS }} }, },
 
 	{"syncer", P_syncer_conf, F_CONFIG_CMD, {{ NULL,
@@ -352,6 +373,13 @@ struct drbd_cmd commands[] = {
 		 { "verify-alg", 'v',T_verify_alg,      ES },
 		 { "cpu-mask",'c',T_cpu_mask,           ES },
 		 { "use-rle",'R',T_use_rle,   EB },
+		 /* deprecated tags are hardcoded here, as they are no longer
+		  * used in current DRBD and would otherwise unnecessarily
+		  * clutter the generated struct definition */
+		 { "delay-probe-volume",'V',	71,	DN(DP_VOLUME,'k',"bytes") },
+		 { "delay-probe-interval",'I',	72,	DN(DP_INTERVAL,1,"1/10 seconds") },
+		 { "throttle-threshold",'T',	73,	DN(RS_THROTTLE_TH,1,"1/10 seconds") },
+		 { "hold-off-threshold",'H',	74,	DN(RS_HOLD_OFF_TH,1,"1/10 seconds") },
 		 CLOSE_OPTIONS }} }, },
 
 	{"new-current-uuid", P_new_c_uuid, F_CONFIG_CMD, {{NULL,
@@ -381,6 +409,7 @@ struct drbd_cmd commands[] = {
 	{"show-gi", P_get_uuids, F_GET_CMD, {.gp={ uuids_scmd} }},
 	{"get-gi", P_get_uuids, F_GET_CMD, {.gp={ uuids_scmd} } },
 	{"show", P_get_config, F_GET_CMD, {.gp={ show_scmd} } },
+	{"check-resize", P_get_config, F_GET_CMD, {.gp={ lk_bdev_scmd} } },
 	{"events",          0, F_EVENTS_CMD, { .ep = {
 		(struct option[]) {
 			{ "unfiltered", no_argument, 0, 'u' },
@@ -457,6 +486,7 @@ static const char *error_messages[] = {
 	EM(ERR_CSUMS_ALG_ND) = "CSUMSAlgNotDigest",
 	EM(ERR_CSUMS_RESYNC_RUNNING) = "Can not change csums-alg while resync is in progress",
 	EM(ERR_PERM) = "Permission denied. CAP_SYS_ADMIN necessary",
+	EM(ERR_NEED_APV_93) = "Protocol version 93 required to use --assume-clean",
 };
 #define MAX_ERROR (sizeof(error_messages)/sizeof(*error_messages))
 const char * error_to_string(int err_no)
@@ -513,7 +543,7 @@ static int dump_tag_list(unsigned short *tlc)
 			break;
 		case TT_STRING:
 			string = (char*)tlc;
-			printf("(string)'%s'",string);
+			printf("(string)'%s'", len ? string : "");
 			break;
 		}
 		printf(" \t[len: %u]\n",len);
@@ -1202,6 +1232,17 @@ static void print_options(struct drbd_option *od, unsigned short *tlc, const cha
 }
 
 
+static void consume_everything(unsigned short *tlc)
+{
+	enum drbd_tags t;
+	int len;
+	while( (t = get_unaligned(tlc)) != TT_END ) {
+		put_unaligned(TT_REMOVED, tlc++);
+		len = get_unaligned(tlc++);
+		tlc = (unsigned short*)((char*)tlc + len);
+	}
+}
+
 static int consume_tag_blob(enum drbd_tags tag, unsigned short *tlc,
 		     char** val, unsigned int* len)
 {
@@ -1438,6 +1479,58 @@ static int show_scmd(struct drbd_cmd *cm, unsigned minor, unsigned short *rtl)
 	return 0;
 }
 
+static int lk_bdev_scmd(struct drbd_cmd *cm, unsigned minor,
+			unsigned short *rtl)
+{
+	struct bdev_info bd = { 0, };
+	char *backing_dev = NULL;
+	uint64_t bd_size;
+	int fd;
+	int idx = idx;
+	int index_valid = 0;
+
+	consume_tag_string(T_backing_dev, rtl, &backing_dev);
+	index_valid = consume_tag_int(T_meta_dev_idx, rtl, &idx);
+
+	/* consume everything */
+	consume_everything(rtl);
+
+	if (!backing_dev) {
+		fprintf(stderr, "Has no disk config, try with drbdmeta.\n");
+		return 1;
+	}
+
+	if (!index_valid) {
+		/* cannot happen, right? ;-) */
+		fprintf(stderr, "No meta data index!?\n");
+		return 1;
+	}
+
+	if (idx >= 0 || idx == DRBD_MD_INDEX_FLEX_EXT) {
+		lk_bdev_delete(minor);
+		return 0;
+	}
+
+	fd = open(backing_dev, O_RDONLY);
+	if (fd == -1) {
+		fprintf(stderr, "Could not open %s: %m.\n", backing_dev);
+		return 1;
+	}
+	bd_size = bdev_size(fd);
+	close(fd);
+
+	if (lk_bdev_load(minor, &bd) == 0 &&
+	    bd.bd_size == bd_size &&
+	    bd.bd_name && !strcmp(bd.bd_name, backing_dev))
+		return 0;	/* nothing changed. */
+
+	bd.bd_size = bd_size;
+	bd.bd_name = backing_dev;
+	lk_bdev_save(minor, &bd);
+
+	return 0;
+}
+
 static int status_xml_scmd(struct drbd_cmd *cm __attribute((unused)),
 		unsigned minor, unsigned short *rtl)
 {
@@ -1606,17 +1699,20 @@ static int uuids_scmd(struct drbd_cmd *cm,
 	       unsigned minor __attribute((unused)),
 	       unsigned short *rtl)
 {
-	uint64_t *uuids;
+	uint64_t uuids[UI_SIZE];
+	char *tl_uuids;
 	int flags = flags;
 	unsigned int len;
 
-	if(!consume_tag_blob(T_uuids,rtl,(char **) &uuids,&len)) {
+	if (!consume_tag_blob(T_uuids, rtl, &tl_uuids, &len)) {
 		fprintf(stderr,"Reply payload did not carry an uuid-tag,\n"
 			"Probably the device has no disk!\n");
 		return 1;
 	}
+
 	consume_tag_int(T_uuids_flags,rtl,&flags);
 	if( len == UI_SIZE * sizeof(uint64_t)) {
+		memcpy(uuids, tl_uuids, len);
 		if(!strcmp(cm->cmd,"show-gi")) {
 			dt_pretty_print_uuids(uuids,flags);
 		} else if(!strcmp(cm->cmd,"get-gi")) {
