@@ -532,6 +532,118 @@ uint64_t bdev_size(int fd)
 	return size64;
 }
 
+char *lk_bdev_path(unsigned minor)
+{
+	char *path;
+	m_asprintf(&path, "%s/drbd-minor-%d.lkbd", DRBD_LIB_DIR, minor);
+	return path;
+}
+
+/* If the lower level device is resized,
+ * and DRBD did not move its "internal" meta data in time,
+ * the next time we try to attach, we won't find our meta data.
+ *
+ * Some helpers for storing and retrieving "last known"
+ * information, to be able to find it regardless,
+ * without scanning the full device for magic numbers.
+ */
+
+/* these return 0 on sucess, error code if something goes wrong. */
+
+/* NOTE: file format for now:
+ * one line, starting with size in byte, followed by tab,
+ * followed by device name, followed by newline. */
+
+int lk_bdev_save(const unsigned minor, const struct bdev_info *bd)
+{
+	FILE *fp;
+	char *path = lk_bdev_path(minor);
+	int ok = 0;
+
+	fp = fopen(path, "w");
+	if (!fp)
+		goto fail;
+
+	ok = fprintf(fp, "%llu\t%s\n",
+		(unsigned long long) bd->bd_size, bd->bd_name);
+	if (ok <= 0)
+		goto fail;
+	if (bd->bd_uuid)
+		fprintf(fp, "uuid:\t"X64(016)"\n", bd->bd_uuid);
+	ok =       0 == fflush(fp);
+	ok = ok && 0 == fsync(fileno(fp));
+	ok = ok && 0 == fclose(fp);
+
+	if (!ok)
+fail:		/* MAYBE: unlink. But maybe partial info is better than no info? */
+		fprintf(stderr, "lk_bdev_save(%s) failed: %m\n", path);
+
+	free(path);
+	return ok <= 0 ? -1 : 0;
+}
+
+/* we may want to remove all stored information */
+int lk_bdev_delete(const unsigned minor)
+{
+	char *path = lk_bdev_path(minor);
+	int rc = unlink(path);
+	if (rc && errno != ENOENT)
+		fprintf(stderr, "lk_bdev_delete(%s) failed: %m\n", path);
+	free(path);
+	return rc;
+}
+
+/* load info from that file.
+ * caller should free(bd->bd_name) once it is no longer needed. */
+int lk_bdev_load(const unsigned minor, struct bdev_info *bd)
+{
+	FILE *fp;
+	char *path;
+	char *bd_name;
+	unsigned long long bd_size;
+	unsigned long long bd_uuid;
+	char nl[2];
+	int rc = -1;
+
+	if (!bd)
+		return -1;
+
+	path = lk_bdev_path(minor);
+	fp = fopen(path, "r");
+	if (!fp) {
+		if (errno != ENOENT)
+			fprintf(stderr, "lk_bdev_load(%s) failed: %m\n", path);
+		goto out;
+	}
+
+	/* GNU format extension: %as:
+	 * malloc buffer space for the resulting char */
+	rc = fscanf(fp, "%llu %as%[\n]uuid: %llx%[\n]",
+			&bd_size, &bd_name, nl,
+			&bd_uuid, nl);
+	/* rc == 5: successfully converted two lines.
+	 *    == 4: newline not found, possibly truncated uuid
+	 *    == 3: first line complete, uuid missing.
+	 *    == 2: new line not found, possibly truncated pathname,
+	 *          or early whitespace
+	 *    == 1: found some number, but no more.
+	 *          incomplete file? try anyways.
+	 */
+	bd->bd_uuid = (rc >= 4) ? bd_uuid : 0;
+	bd->bd_name = (rc >= 2) ? bd_name : NULL;
+	bd->bd_size = (rc >= 1) ? bd_size : 0;
+	if (rc < 1) {
+		fprintf(stderr, "lk_bdev_load(%s): parse error\n", path);
+		rc = -1;
+	} else
+		rc = 0;
+
+	fclose(fp);
+out:
+	free(path);
+	return rc;
+}
+
 void get_random_bytes(void* buffer, int len)
 {
 	int fd;
@@ -597,3 +709,62 @@ int m_asprintf(char **strp, const char *fmt, ...)
 	return r;
 }
 
+/* print len bytes from buf in the format of well known "hd",
+ * adjust displayed offset by file_offset */
+void fprintf_hex(FILE *fp, off_t file_offset, const void *buf, unsigned len)
+{
+	const unsigned char *c = buf;
+	unsigned o;
+	int skipped = 0;
+
+	for (o = 0; o + 16 < len; o += 16, c += 16) {
+		if (o && !memcmp(c - 16, c, 16)) {
+			skipped = 1;
+			continue;
+		}
+		if (skipped) {
+			skipped = 0;
+			fprintf(fp, "*\n");
+		}
+		/* no error check here, don't know what to do about errors */
+		fprintf(fp,
+			/* offset */
+			"%08llx"
+			/* two times 8 byte as byte stream, on disk order */
+			"  %02x %02x %02x %02x %02x %02x %02x %02x"
+			"  %02x %02x %02x %02x %02x %02x %02x %02x"
+			/* the same as printable char or '.' */
+			"  |%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c|\n",
+			(unsigned long long)o + file_offset,
+			c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7],
+			c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15],
+
+#define p_(x)	(isprint(x) ? x : '.')
+#define p(a,b,c,d,e,f,g,h) \
+		p_(a), p_(b), p_(c), p_(d), p_(e), p_(f), p_(g), p_(h)
+			p(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]),
+			p(c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15])
+		       );
+	}
+	if (skipped) {
+		skipped = 0;
+		fprintf(fp, "*\n");
+	}
+	if (o < len) {
+		unsigned remaining = len - o;
+		unsigned i;
+		fprintf(fp, "%08llx ", (unsigned long long)o + file_offset);
+		for (i = 0; i < remaining; i++) {
+			if (i == 8)
+				fprintf(fp, " ");
+			fprintf(fp, " %02x", c[i]);
+		}
+		fprintf(fp, "%*s  |", (16 - i)*3 + (i < 8), "");
+		for (i = 0; i < remaining; i++)
+			fprintf(fp, "%c", p_(c[i]));
+#undef p
+#undef p_
+		fprintf(fp, "|\n");
+	}
+	fprintf(fp, "%08llx\n", (unsigned long long)len + file_offset);
+}
